import os
import json
import re
import warnings
import datetime as dt

import h5py
import numpy as np
from scipy.constants import physical_constants

import exceptions

class CXIFile():
    """A ptychography CXI file.
    """
    def __init__(self, path, mode='r'):
        self.path = path
        self.filename = os.path.basename(path)
        self.hdf_file = h5py.File(path, mode=mode)
    
    def __repr__(self):
        return '<CXI: {}>'.format(self.filename)
    
    def __str__(self):
        return self.filename
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
    
    def close(self):
        self.hdf_file.close()
    
    def filenames(self):
        filenames = [os.path.basename(self.path)]
        return filenames
    
    def image_frames(self):
        data = self.hdf_file['/entry_1/image_1/data'].value[::-1]
        return np.array([data])
    
    def image_shape(self):
        shape = self.hdf_file['/entry_1/image_1/data'].shape
        return shape
    
    def num_images(self):
        return 1
    
    def energies(self):
        energy = self.hdf_file['/entry_1/instrument_1/source_1/energy'].value
        energy = energy / physical_constants['electron volt'][0]
        return [energy]
    
    def pixel_size(self):
        try:
            px_size = float(self.hdf_file['/entry_1/process_1/Param/pixnm'].value)
        except KeyError:
            raise exceptions.DataFormatError(
                'File {} does not contain pixel size'
                ''.format(self.filename))
        return px_size


class HDRFile():
    """A STXM HDR file."""
    def __init__(self, path):
        self.path = path
        self.filename = os.path.basename(path)
        self.xim_fmt = '{base}_a{idx}.xim'.format(
            base=os.path.splitext(self.path)[0], idx='{idx:03d}')
        with open(self.path, mode='r') as f:
            self._contents = f.read()
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        pass
    
    def close(self):
        pass
    
    def filenames(self):
        # Extract the file IDs from the image metadata
        ids = (int(s['id'].split('_')[0]) for s in self._image_metadata())
        # Convert the image IDs to filenames
        fnames = [self.xim_fmt.format(idx=i) for i in ids]
        return fnames
    
    def image_shape(self):
        shape_re = re.compile(r'XPoints = (\d+); YPoints = (\d+)')
        match = shape_re.search(self._contents)
        if not match:
            raise exceptions.DataFormatError('Could not read image_shape in'
                                             '{}'.format(self.path))
        else:
            shape = tuple(int(match.group(i)) for i in (1, 2))
        return shape
    
    def energies(self):
        regex = (
            r'\s*StackAxis = { Name = "Energy"; Unit = "eV";'
            r' Min = (\d+); Max = (\d+); Dir = (\d+);\n'
            r'\s*Points = \((\d+), ([0-9., ]+)\);\n'
            r'};'
        )
        E_re = re.compile(regex)
        match = E_re.search(self._contents)
        # Check if a valid match was found
        if not match:
            raise exceptions.DataFormatError('Could not read energies in'
                                             '{}'.format(self.path))
        else:
            # Decode the groups from the regular expression
            num_energies = int(match.group(4))
            energies = [float(e) for e in match.group(5).split(',')]
            if num_energies != len(energies):
                log.warning('Wrong number of energies in "{}": expected {} got {}'
                            ''.format(self.path, num_energies, len(energies)))
        # Check that the number of energies matches the number of images
        num_images = self.num_images()
        if num_energies != num_images:
            log.warning('Number of energies ({}) does not match '
                        'number of images ({}) in {}'
                        ''.format(num_energies, num_images, self.filename))
        return energies
    
    def num_images(self):
        return len(self._image_metadata())
    
    def image_frames(self):
        frames = np.array([np.loadtxt(xim, dtype='int')
                           for xim in self.filenames()])
        return frames
    
    def _image_metadata(self):
        """Get a list of dictionaries with metadata for each image frame."""
        regex = re.compile(
            r'Image(?P<id>[0-9_]+) = {'
            r'StorageRingCurrent = (?P<current>[-0-9.]+); '
            r'Energy = (?P<energy>[-0-9.]+); '
            r'Time = "(?P<time>[0-9a-zA-Z :]+)"; '
            r'ZP_dest= (?P<zp_dest>[-0-9.]+); '
            r'ZP_error = (?P<zp_error>[-0-9.]+);};'
        )
        matches = tuple(m.groupdict() for m in regex.finditer(self._contents))
        # Convert strings to numbers, dates, etc.
        types = {
            'energy': float,
            'time': lambda t: dt.datetime.strptime(t, '%Y %b %d %H:%M:%S'),
            'current': float,
            'zp_dest': float,
            'zp_error': float,
        }
        for scan in matches:
            for key, val in scan.items():
                type_ = types.get(key, lambda x: x)
                scan[key] = type_(val)
        return matches
    
    def scan_metadata(self):
        regex = (
            r'\{\s*CentreXPos = (?P<centerx>-?\d+\.?\d*);'
            r'\s*CentreYPos = (?P<centery>-?\d+\.?\d*);'
            r'\s*XRange = (?P<xrange>\d+\.?\d*);'
            r'\s*YRange = (?P<yrange>\d+\.?\d*);'
            r'\s*XStep = (?P<xstep>\d+\.?\d*);'
            r'\s*YStep = (?P<ystep>\d+\.?\d*);'
            r'\s*XPoints = (?P<xpoints>\d+\.?\d*);'
            r'\s*YPoints = (?P<ypoints>\d+\.?\d*);'
            r'\s*SquareRegion = (?P<squareregion>[a-z]+);'
            r'\s*SquarePixels = (?P<squarepixels>[a-z]+);'
        )
        scan_re = re.compile(regex)
        match = scan_re.search(self._contents)
        if match:
            metadata = match.groupdict()            
        else:
            raise exceptions.DataFormatError('Cannot read format in file {}.'
                                             ''.format(self.filename))
        # Convert value to proper datatypes
        to_bool = lambda x: True if x == 'true' else False
        types = {
            'centerx': float,
            'centery': float,
            'xrange': float,
            'yrange': float,
            'xstep': float,
            'ystep': float,
            'xpoints': int,
            'ypoints': int,
            'squareregion': to_bool,
            'squarepixels': to_bool,
            
        }
        for key, val in metadata.items():
            type_ = types.get(key, lambda x: x)
            metadata[key] = type_(val)
        return metadata
    
    def pixel_size(self):
        """Retrieve the nominal pixel size in nm."""
        metadata = self.scan_metadata()
        xstep, ystep = metadata['xstep'], metadata['ystep']
        # Convert values from µm to nm
        xstep, ystep = xstep * 1000, ystep * 1000
        # Check for non-square scans
        if xstep != ystep:
            warnings.warn('Mismatched x ({}) and y ({}) steps in scan {}. '
                          'Ignoring y step.'.format(xstep, ystep, self.filename))
        # Just use the xstep as the step size
        return xstep



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for setting up and displaying plots using matplotlib."""

from contextlib import contextmanager
from typing import List, NoReturn

import numpy as np
from matplotlib import pyplot, cm, rcParams, rc_context, style
from matplotlib.colors import Normalize
from matplotlib.ticker import ScalarFormatter

from utilities import pixel_to_xy, Extent, Pixel

@contextmanager
def latexify(styles: List[str]=[], preamble: List[str]=[]):
    """Set some custom options for saving matplotlib graphics in PGF
    format.
    
    Use this as a context manager, along with additional matplotlib styles:
    
    .. code:: python
        
        with xp.latexify(['beamer']):
            plt.plot(...)
            
    
    This will let you add in LaTeX tools and mpl styles together. By
    default, ``siunitx`` and ``mhchem`` packages are
    included. Additional ``\\usepackage`` statements can be included
    using the ``preamble`` parameter.
    
    Parameters
    ==========
    styles : optional
      Additional matplotlib styles in load in the context.
    preamble : optional
      Additional lines to add to the LaTeX preamble.
    
    """
    # Set default LaTeX PGF style
    pgf_with_latex = {                      # setup matplotlib to use latex for output# {{{
        "pgf.texsystem": "xelatex",        # change this if using xetex or lautex
        # "font.family": "serif",
        "font.serif": [],                   # blank entries should cause plots 
        "font.sans-serif": [],              # to inherit fonts from the document
        "font.monospace": [],
        "pgf.preamble": [
            r"\usepackage[utf8x]{inputenc}",    # use utf8 fonts 
            r"\usepackage[T1]{fontenc}",        # plots will be generated
            r"\usepackage{fontspec}",
            r"\usepackage[detect-all,locale=DE,per-mode=reciprocal]{siunitx}",
            r"\usepackage[version=4]{mhchem}",
        ] + preamble,
    }
    # Enter the context library
    with rc_context(rc=pgf_with_latex):
        style.use(styles)
        yield


def remove_extra_spines(ax):
    """Removes the right and top borders from the axes."""
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.yaxis.set_ticks_position('left')
    ax.xaxis.set_ticks_position('bottom')
    return ax


def set_axes_color(ax, color):  # pragma: no cover
    """Set the axes, tick marks, etc of `ax` to mpl color `color`. Also,
    "doegreen" has special significance as the color associated with the
    US department of energy."""
    if color == "doegreen":
        color = (33/255, 99/255, 50/255)
    # Set the spine colors
    ax.spines['bottom'].set_color(color)
    ax.spines['top'].set_color(color)
    ax.spines['right'].set_color(color)
    ax.spines['left'].set_color(color)
    # Change ticks colors
    ax.tick_params(axis='x', colors=color)
    ax.tick_params(axis='y', colors=color)
    # Change label color
    ax.yaxis.label.set_color(color)
    ax.xaxis.label.set_color(color)


def set_outside_ticks(ax):  # pragma: no cover
    """Convert all the axes so that the ticks are on the outside and don't
    obscure data."""
    ax.get_yaxis().set_tick_params(which='both', direction='out')
    ax.get_xaxis().set_tick_params(which='both', direction='out')
    return ax


def new_axes(height=5, width=None):
    """Create a new set of matplotlib axes for plotting. Height in inches."""
    # Adjust width to accomodate colorbar
    if width is None:
        width = height / 0.8
    fig = pyplot.figure(figsize=(width, height))
    # Set background to be transparent
    fig.patch.set_alpha(0)
    # Create axes
    ax = pyplot.gca()
    # Remove borders
    remove_extra_spines(ax)
    return ax


def new_image_axes(height=5, width=5):  # pragma: no cover
    """Square axes with ticks on the outside."""
    ax = new_axes(height, width)
    return set_outside_ticks(ax)


def big_axes():  # pragma: no cover
    """Return a new Axes object, but larger than the default."""
    return new_axes(height=9, width=16)


def dual_axes(fig=None, longdim=13.8, shortdim=6.9, orientation='horizontal'):  # pragma: no cover
    """Two new axes for mapping, side-by-side.

    Parameters
    ----------
    longdim : float
      Size in inches for the long dimension. If orientation is
      "vertical", this will be the height.
    shortdim : float
      Size in inches for the short dimension. If orientation is
      "vertical", this will be the width.
    """
    if fig is None:
        fig = pyplot.figure()
    if orientation == 'vertical':
        fig, (ax1, ax2) = fig.subplots(2, 1)
        fig.set_figwidth(shortdim)
        fig.set_figheight(longdim)
    else:
        fig, (ax1, ax2) = pyplot.subplots(1, 2)
        fig.set_figwidth(longdim)
        fig.set_figheight(shortdim)
    # Remove redundant borders
    remove_extra_spines(ax1)
    remove_extra_spines(ax2)
    # Set background to be transparent
    fig.patch.set_alpha(0)
    return (ax1, ax2)


def draw_histogram_colorbar(ax, *args, **kwargs):  # pragma: no cover
    """Similar to `draw_colorbar()` with some special formatting options
    to put it along the X-axis of the axes."""
    cbar = draw_colorbar(ax=ax, pad=0, orientation="horizontal", energies=None, *args, **kwargs)
    ax.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='off',
        labeltop='off')
    ax.spines['bottom'].set_visible(False)
    cbar.ax.set_xlabel(ax.get_xlabel())
    ax.xaxis.set_visible(False)
    cbar.ax.set_title("")
    cbar.outline.set_visible(False)
    gray = (0.1, 0.1, 0.1)
    cbar.ax.axhline(cbar.ax.get_ylim()[1], linewidth=2, linestyle=":", color=gray)
    cbar.ax.tick_params(
        axis='x',
        which='both',
        bottom='on',
        top='on',
        labelbottom="on",
    )
    return cbar


def draw_colorbar(ax, cmap, norm, energies=None, orientation="vertical",
                  *args, **kwargs):  # pragma: no cover
    """Draw a colorbar on the side of a mapping axes to show the range of
    colors used. Returns the newly created colorbar object.

    Parameters
    ---------
    ax
      Matplotlib axes object against which to plot.
    cmap : str
      String or mpl Colormap instance indicating which colormap to
      use.
    norm : matplotlib.Normalize
      Describes the range of values to use.
    energies : iterable, optional
      Values to put as the tick marks on the colorbar. If not given, 3
      points across ``norm`` will be used.
    orientation : str, optional
      "horizontal" or "vertical" (default)
    
    """
    mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
    mappable.set_array(np.arange(0, 3))
    # Prepare default energies if necessary
    if energies is None:
        energies = np.linspace(norm.vmin, norm.vmax, num=3)
    # Add the colorbar to the axes
    cbar = pyplot.colorbar(mappable,
                           ax=ax,
                           ticks=energies,
                           spacing="proportional",
                           orientation=orientation,
                           *args, **kwargs)
    # Annotate the colorbar
    cbar.ax.set_title('eV')
    # Make sure the ticks don't use scientific notation
    cbar.formatter.set_useOffset(False)
    cbar.update_ticks()
    return cbar


def plot_pixel_spectra(pixels, extent, spectra, energies, map_ax,
                       spectra_ax=None, step_size=0):  # pragma: no cover
    """Highlight certain pixels in an already-plotted map and plot their
    spectra. The map should already have been plotted.

    Arguments
    ---------

    - pixels : List of 2-tuples with (row, column) positions of pixels
      to highlight and plot.

    - extent : Matplotlib extent specification for converting (y, x)
      positions to (row, column) positions.

    - spectra : Numpy array with spectra that will be plotted. Entries
      in `pixels` will use this array to get spectra. Shape is assumed
      to be (row, column, energy).

    - energies : Numpy array with energies to use as the x-axis for
      plotting spectra.

    - map_ax : The matplotlib Axes that will be used for highlighting
      pixels.

    - spectra_ax : A matplotlib Axes on which to plot the spectra. If
      None (default) a new axes will be created.

    - step_size : An offset to add to each spectrum so they are not
      directly on top of each other.
    """
    # Prepare some metadata
    map_shape = spectra.shape[0:2]
    extent = Extent(*extent)
    # Prepare axes if necessary (and remove spines for arbitrary units)
    if spectra_ax is None:
        spectra_ax = new_axes()
    remove_extra_spines(spectra_ax)
    # Remove ticks on the y axes
    tick_params = {
        'axis': 'y',
        'which': 'both',
        'left': 'off',
        'right': 'off',
        'labelright': 'off',
        'labelleft': 'off',
    }
    # Prepare a secondary y axes if necessary for complex data
    if np.iscomplexobj(spectra):
        spectra_ax2 = spectra_ax.twinx()
        remove_extra_spines(spectra_ax2)
        spectra_ax2.tick_params(**tick_params)
    spectra_ax.tick_params(**tick_params)
    # Do the plotting and annotating for each pixel
    min_val = 0
    max_val = 0
    for idx, px in enumerate(pixels):
        # Put an annotation on the map
        px = Pixel(*px)
        xy = pixel_to_xy(px, extent, shape=map_shape)
        map_ax.text(xy.x, xy.y, str(idx), color="white",
                    horizontalalignment="center", verticalalignment="center",)
        # Normalize the spectrum
        spectrum = spectra[px]
        min_ = np.min(np.abs(spectrum))
        max_ = np.max(np.abs(spectrum))
        spectrum = (spectrum - min_) / (max_ - min_)
        # Plot the spectrum on the spectrum axes
        spectrum += step_size * idx
        plot_spectrum(spectrum, energies, ax=spectra_ax,
                      ax2=spectra_ax2)
        # Put labels on the axes to indicate which specturm is which
        spectra_ax.text(
            x=np.max(energies) + 2,
            y=np.mean((np.real(spectrum), np.imag(spectrum))),
            s=idx,
        )
        # Save values for proper scaling of axes
        min_val = complex(
            min(np.real(min_val), np.real(spectrum).min()),
            min(np.imag(min_val), np.imag(spectrum).min())
        )
        max_val = complex(
            max(np.real(max_val), np.real(spectrum).max()),
            max(np.imag(max_val), np.imag(spectrum).max())
        )
    # Set axes limits based on min and max values
    min_val -= min_val * 0.1
    max_val += max_val * 0.1
    spectra_ax.set_ylim(bottom=np.real(min_val), top=np.real(max_val))
    if spectra_ax2 is not None:
        spectra_ax2.set_ylim(bottom=np.imag(min_val), top=np.imag(max_val))
    spectra_ax.set_xlim(right=np.max(energies)+4)

def plot_kedge_fit(energies, params):
    """Plot the fit based on the given k-edge params. The params will be
    given to ``xanespy.xanes_math.KEdgeParams`` for conversion.

    energies : np.ndarray
      A 1D array with the x values to plot.

    params : array-like
      Fitted K-edge parameters.
    """
    p = KEdgeParams(*params)
    # Plot the sigmoid by making everything else zero
    sig_p = dict(zip(kedge_params, p))
    sig_p['ga'] = 0
    sig_p = KEdgeParams(**sig_p)
    sig = predict_edge(energies, *sig_p)
    pyplot.plot(energies, sig, linestyle=":")
    # Plot the whole predicted line
    fit = predict_edge(energies, *p)
    pyplot.plot(energies, fit, linestyle=":")

def plot_spectrum(spectrum, energies, norm=Normalize(),
                  show_fit=False, ax=None, ax2=None,
                  linestyle=':', color="blue", cmap="plasma",
                  polar_coords=False,
                  *args, **kwargs):  # pragma: no cover
    """Plot an energy spectrum on an axes.
    
    Applies some color formatting if `edge` is a valid XANES Edge
    object.
    
    Arguments
    ---------
    spectrum : np.ndarray
      Array of intensity values.
    energies : np.ndarray
      Array of energy values.
    norm : optional
      Matplotlib Normalize() object that shows the map
      range. This will be used to annotate the plot if it is give.
    show_fit : bool, optional
      Whether to plot lines showing the best fit.
    ax : mpl.Axes, optional
      Matplotlib Axes on which to plot. If not given, a new axes
      will be generated.
    ax2 : mpl.Axes, optional
      A second y-axes for plotting the imaginary component if
      the data are complex.
    linestyle : optional
      Passed on to matplotlib.
    cmap : str, optional
      Colormap, passed on to matplotlib
    color : optional
      Specifies the color for the circles plotted. Either "x"
      or "y" will decide based on the numerical value, `norm` and
      `cmap` arguments. Anything else will be passed as a color spec
      to the matplotlib commands.
    polar_coords : bool, optional
      If truthy, the spectrum will be plotted as modulus-phase instead
      of real-imag. For purely real data this is equivalent to taking
      the absolute value.
    
    """
    if ax is None:
        ax = new_axes()
    if norm is not None:
        norm.autoscale_None(np.real(spectrum))
    # Retrieve `values` in case it's a pandas series
    spectrum = getattr(spectrum, 'values', spectrum)
    # Color code the markers by energy
    is_complex = np.iscomplexobj(spectrum)
    if is_complex:
        colors = None
    elif color == "x":
        colors = cm.get_cmap(cmap)(norm(energies))
    elif color == "y":
        colors = cm.get_cmap(cmap)(norm(spectrum))
    else:
        colors = [color] * len(energies)
    # Remove secondary axes and re-add them for complex values
    if is_complex:
        # Check if modulus-phase or real-imag components
        if polar_coords:
            comp0, comp1 = (np.abs, np.angle)
            label0, label1 = ('Modulus', 'Phase')
        else:
            comp0, comp1 = (np.real, np.imag)
            label0, label1 = ('Real', 'Imag')
        # Create a secondary axis
        if ax2 is None:
            ax2 = ax.twinx()
        # Convert complex values to two lines
        ys = spectrum
        artist = ax.plot(energies, comp0(ys), linestyle=linestyle, color="C0")
        artist.extend(ax2.plot(energies, comp1(ys),
                               linestyle=linestyle, color="C1", *args, **kwargs))
    else:
        # Just plot the real numbers
        ys = np.real(spectrum)
        artist = ax.plot(energies, ys, linestyle=linestyle, *args, **kwargs)
    # save limits, since they get messed up by scatter plot
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    # Draw scatter plot of data points
    markersize = kwargs.get('markersize', rcParams['lines.markersize'])
    if is_complex:
        ax.scatter(energies, comp0(ys), c="C0", s=25, alpha=0.5)
        ax.set_ylabel(label0, color="C0")
        for t1 in ax.get_yticklabels():
            t1.set_color("C0")
        ax2.scatter(energies, comp1(ys), c="C1", s=markersize ** 2, alpha=0.5)
        ax2.set_ylabel(label1, color="C1")
        for t1 in ax2.get_yticklabels():
            t1.set_color("C1")
    else:
        ax.scatter(energies, ys, c=colors, s=markersize ** 2)
        ax.set_ylabel('Optical Depth')
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    ax.set_xlabel('Energy /eV')
    # Plot the edges of the map range
    if norm is not None:
        vlineargs = dict(linestyle='--', alpha=0.7,
                         color="lightgrey", zorder=0,)
        ax.axvline(norm.vmin, **vlineargs)
        ax.axvline(norm.vmax, **vlineargs)
    return artist


def plot_composite_map(data, ax=None, origin="upper", *args, **kwargs):  # pragma: no cover
    """Plot an RGB composite map on the given axes."""
    while data.shape[-1] < 3:
        data = np.concatenate((data, np.zeros_like(data)),
                              axis=-1)
    data = data[..., 0:3]
    if ax is None:
        ax = new_image_axes()
    artist = ax.imshow(data, origin=origin, *args, **kwargs)
    return artist


def plot_txm_map(data, edge=None, norm=None, ax=None, cmap='plasma',
                 origin="upper", vmin=None, vmax=None,
                 *args, **kwargs):  # pragma: no cover
    # Get a default normalizer
    if norm is None:
        norm = Normalize(vmin, vmax)
    norm.autoscale_None(data[~np.isnan(data)])
    # Create axes if necessary
    if ax is None:
        ax = new_image_axes()
        if edge is None:
            # No specific edge is given, so use up to 10 evenly space energies
            num_Es = round(min(abs(norm.vmax - norm.vmin + 1), 10))
            energies = np.linspace(norm.vmin, norm.vmax, num=num_Es)
        else:
            energies = edge.energies_in_range(norm_range=(norm.vmin, norm.vmax))
        draw_colorbar(ax=ax, cmap=cmap, norm=norm, energies=energies)
    # Do the plotting
    artist = ax.imshow(data,
                       cmap=cmap,
                       origin=origin,
                       norm=norm,
                       *args, **kwargs)
    # Add annotations and formatting stuff
    return artist


def plot_txm_histogram(data, ax=None, norm=None, bins=None,
                       cmap='plasma', add_cbar=True, *args, **kwargs):  # pragma: no cover
    """Take an array of data values and show a histogram with some
    color-coding related to normalization value.

    Returns: The matplotlib axes object used for plotting.

    Parameters
    ----------
    data : np.ndarray
      An array of values to plot on the histogram.
    ax : optional
      Matplotlib Axes instance to receive the plot. If None, a new axes
      will created.
    norm : optional
      Matplotlib Normalize instance with the colormap range.
    bins : optional
      Bins to pass to the matplotlib hist() routine. If None
      (default), we will choose based on dtype of the data: integers
      will yield 1-wide bins, anything else will give 256 bins.
    cmap : str, optional
      Matplotlib colormap for coloring the bars.
    add_cbar : bool, optional
      Boolean to decide whether to add a colorbar along the
      bottom axis or not.
    *args
      Positional arguments passed to matplotlib's `hist` call.
    *kwargs
      Keyword arguments passed to matplotlib's `hist` call.
    
    """
    if ax is None:
        ax = new_axes()
    # Flatten the data so it can be nicely plotted
    if hasattr(data, 'mask'):
        data = data[~data.mask]
    else:
        data = data.flatten()
    # Remove any Not-a-number values
    data = data[~np.isnan(data)]
    # Set normalizer
    if norm is None:
        norm = Normalize()
        norm.autoscale_None(data)
    # Clip the data so that it includes the end-members
    clip_data = np.clip(data, norm.vmin, norm.vmax)
    unique_vals = np.unique(np.array(clip_data))
    # Determine a reasonable binning parameter based on data type
    if bins is not None:
        pass
    elif np.issubdtype(data.dtype, np.integer):
        # integers, so make the bins 1-wide
        bins = np.arange(norm.vmin, norm.vmax, step=1)
    elif len(unique_vals) < 256:
        # Not many values to choose from, so use the values as bins
        bins = np.sort(unique_vals)
    else:
        bins = 256
    # Plot the histogram
    n, bins, patches = ax.hist(clip_data, bins=bins, *args, **kwargs)
    # Set colors on histogram
    for patch in patches:
        x_position = patch.get_x()
        real_cmap = cm.get_cmap(cmap)
        color = real_cmap(norm(x_position))
        patch.set_color(color)
    # Set axes decorations
    ax.set_ylabel("Pixels")
    ax.set_xlim(norm.vmin, norm.vmax)
    # ax.xaxis.set_ticks(energies)
    ax.xaxis.get_major_formatter().set_useOffset(False)
    # Draw a colorbar along the bottom axes
    if add_cbar:
        cbar = draw_histogram_colorbar(ax=ax, cmap=cmap, norm=norm)
    else:
        cbar = None
    return (ax, cbar)


def plot_spectra(spectra, energies, ax=None):
    """Take an iterable of spectra and plot them one above the next.
    
    Parameters
    ----------
    spectra : iterable
      Each entry is an array with intensity values.
    energies : np.ndarray
      Energy values for the points in each entry of ``spectra``.
    ax : matplotlib.Axes, optional
      The Axes object to receive the plots. If omitted, a new Axes
      will be created.
    
    Returns
    -------
    artists : list
      A list of the matplotlib artists used to draw the spectra.
    
    """
    artists = []
    # Get a default axes if one was not provided
    if ax is None:
        ax = new_axes()
    # Do the plotting for each spectrum
    yoffset = 0
    for spectrum in spectra:
        artist = ax.plot(energies, spectrum + yoffset, marker='o', linestyle=':')
        artists.append(artist)
        # Calculate the appropriate step for the next round
        yoffset += np.median(spectrum) - 0.25 * (np.max(spectrum) - np.min(spectrum))
    return artists


def plot_spectra_as_map(spectra, energies, ax=None, extent=None, **kwargs):
    """Take an iterable of spectra and plot them as a heat map.
    
    This function takes energies so it can put ticks on the x-axis,
    however if they are not equally spaced, their distance on the map
    will not properly capture they're distance numerically.
    
    Parameters
    ----------
    spectra : iterable
      Each entry is an array with intensity values.
    energies : np.ndarray
      Energy values for the points in each entry of ``spectra``.
    ax : matplotlib.Axes, optional
      The Axes object to receive the plots. If omitted, a new Axes
      will be created.
    extent : Matplotlib extent.
    **kwargs : Get passed on to matplotlib imshow().
    
    Returns
    -------
    artists : list
      A list of the matplotlib artists used to draw the spectra.
    
    """
    # Get a default axes if one was not provided
    if ax is None:
        ax = new_axes()
    # Do the plotting for each spectrum
    artist = ax.imshow(spectra, origin='lower', extent=extent, aspect="auto", **kwargs)
    return [artist]


def plot_txm_intermediates(images):  # pragma: no cover
    """Accept a dictionary of images and plots them each on its own axes
    using matplotlib's `imshow`. This is a complement to routines that
    operate on a microscopy frame and optionally return all the
    intermediate calculated frames.
    
    """
    for key in images.keys():
        ax1, ax2 = dual_axes()
        ax1.imshow(images[key], cmap='gray')
        ax1.set_title(key)
        ax2.hist(images[key].flat, bins=100)



# -*- coding: utf-8 -*-
#
# Copyright © 2016 Mark Wolf
#
# This file is part of Xanespy.
#
# Xanespy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Xanespy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Xanespy. If not, see <http://www.gnu.org/licenses/>.

"""Tools for accessing TXM data stored in an HDF5 file."""

import logging
from collections import namedtuple

import h5py
import numpy as np
from tqdm import tqdm

from exceptions import (GroupKeyError, CreateGroupError, FrameSourceError,)
import xanes_math as xm
from utilities import get_component


log = logging.getLogger(__name__)


def merge_stores(base_store, new_store, destination, energy_difference=0.25, upsample=True):
    """Merge two open txm stores into a third store.
    
    Framesets will be combined from both ``base_store`` and
    ``new_store``. If frames in both sets are within
    ``energy_difference`` or each other, then the one from
    ``new_store`` will be used. The resulting frames will be cropped
    and up-sampled. Maps will not be copied, since they are unlikely
    to be reliable with the merged framesets. The metadata will
    reflect the merging as best as possible.
    
    """
    num_timesteps = len(base_store.timestep_names)
    M = namedtuple('M', ('energy', 'idx', 'store'))
    # Create some arrays to hold the results
    energies = []
    intensities = []
    optical_depths = []
    filenames = []
    pixel_sizes = []
    for t_idx in range(num_timesteps):
        # Prepare a "master list" of frames to used
        num_energies_A = len(base_store.energies[t_idx])
        energies_A = base_store.energies[t_idx]
        master_list = [M(energies_A[E_idx], E_idx, base_store)
                       for E_idx in range(num_energies_A)]
        # Update the master list for each frame in the new store
        energies_B = new_store.energies[t_idx]
        for E_idx, E in enumerate(energies_B):
            # Check if this entry exists in the master list already
            matches = [m for m in master_list
                       if abs(m.energy - E) < energy_difference]
            if matches:
                # It already exists, so replace it
                for m in matches:
                    idx = master_list.index(m)
                    master_list[idx] = M(E, E_idx, new_store)
            else:
                # It doesn't exist, so add it
                master_list.append(M(E, E_idx, new_store))
        # Sort the master list to be in energy ascending order
        master_list.sort(key=lambda m: m.energy)
        # Empty array for catching processed data
        Es = []
        Is = []
        ODs = []
        fnames = []
        # Prepare the arguments for resizing each image
        shapes = [m.store.intensities.shape[2:] for m in master_list]
        max_shape = (max(s[0] for s in shapes),
                     max(s[1] for s in shapes))
        dims = [np.array(m.store.intensities.shape[2:]) * m.store.pixel_sizes[t_idx,E_idx]
                for m in master_list]
        min_dims = (min(d[0] for d in dims), min(d[1] for d in dims))
        target_px_size = min(m.store.pixel_sizes[t_idx,E_idx] for m in master_list)
        pixel_sizes.append((target_px_size,) * len(master_list))
        # Retrieve and resize each image
        for m, dim in tqdm(zip(master_list, dims), total=len(master_list)):
            Es.append(m.energy)
            px_size = m.store.pixel_sizes[t_idx]
            I = m.store.intensities[t_idx,m.idx]
            if np.iscomplexobj(I):
                comp = 'imag'
            else:
                comp = 'real'
            I = get_component(I, comp)
            I = xm.resample_image(I, new_shape=max_shape, src_dims=dim, new_dims=min_dims)
            Is.append(I)
            OD = m.store.optical_depths[t_idx,m.idx]
            OD = get_component(OD, comp)
            OD = xm.resample_image(OD, new_shape=max_shape, src_dims=dim, new_dims=min_dims)
            ODs.append(OD)
            # Save the necessary metadata
            fnames.append(m.store.filenames[t_idx][m.idx])
        # Save the combined framesets
        energies.append(Es)
        intensities.append(Is)
        optical_depths.append(ODs)
        filenames.append(fnames)
    # Set the newly merged frames
    destination.energies = energies
    destination.intensities = np.array(intensities)
    destination.optical_depths = np.array(optical_depths)
    destination.filenames = filenames
    destination.timestep_names = base_store.timestep_names
    destination.pixel_sizes = pixel_sizes
    destination.pixel_unit = base_store.pixel_unit


class TXMDataset():
    """Data descriptor for accessing HDF datasets.
    
    Parameters
    ----------
    name : str
      The dataset name in the HDF file.
    context : str, optional
      Type of dataset this is: frameset, map, metadata, etc.
    dtype : np.dtype, optional
      The data-type to use when saving new data to disk. Using lower
      precision datatypes can save significant disk space.
    
    """
    def __init__(self, name, context=None, dtype=None):
        self.name = name
        self.context = context
        self.dtype = dtype
    
    def __get__(self, store, type=None):
        dataset = store.get_dataset(self.name)
        return dataset
    
    def __set__(self, store, value):
        store.replace_dataset(name=self.name, data=value,
                              context=self.context, dtype=self.dtype)
    
    def __delete__(self, store):
        del store.data_group()[self.name]


class TXMStore():
    """Wrapper around HDF5 file that stores TXM data.
    
    It has a series of descriptors and properties that return the
    corresponding HDF5 dataset object; the TXMStore().attribute.value
    pattern can be used to get pure numpy arrays directly. These
    objects should be used as a context manager to ensure that the
    file is closed, especially if using a writing mode:
    
        with TXMStore() as store:
            # Do stuff with store here
    
    Parameters
    ----------
    hdf_filename : str
      Path to the HDF file to be used.
    parent_name : str
      Name of the top-level HDF5 group.
    data_name : str
      Name of the second level HDF5 group, used for specific data
      iterations (eg. imported, aligned)
    mode : str
      Eg. 'r' for read-only, 'r+' for read-write. Passed directly to
      h5py.File constructor.
    
    """
    VERSION = 1
    _data_name = None
    
    # HDF5 Descriptors
    # ----------------
    energies = TXMDataset('energies', context='metadata')
    signals = TXMDataset('extracted_signals', context='metadata')
    original_positions = TXMDataset('original_positions', context='metadata')
    timestep_names = TXMDataset('timestep_names', context='metadata')
    pixel_sizes = TXMDataset('pixel_sizes', context='metadata')
    linear_combination_sources = TXMDataset('linear_combination_sources', context='metadata')
    
    intensities = TXMDataset('intensities', context='frameset')
    optical_depths = TXMDataset('optical_depths', context='frameset')
    references = TXMDataset('references', context='frameset')
    signal_weights = TXMDataset('extracted_signal_weights', context='frameset')
    linear_combination_parameters = TXMDataset('linear_combination_parameters', context='frameset')
    
    optical_depth_mean = TXMDataset('optical_depth_mean', context='map')
    intensity_mean = TXMDataset('intensity_mean', context='map')
    signal_map = TXMDataset('signal_map', context='map')
    edge_mask = TXMDataset('edge_mask', context='map')
    whiteline_max = TXMDataset('whiteline_max', context='map')
    whiteline_fit = TXMDataset('whiteline_fit', context='map')
    cluster_fit = TXMDataset('cluster_fit', context='map')
    particle_labels = TXMDataset('particle_labels', context='map')
    segments = TXMDataset('segments', context='map')
    linear_combination_residuals = TXMDataset('linear_combination_residuals', context='map')
    
    def __init__(self, hdf_filename: str,
                 parent_name: str, data_name=None,
                 mode='r'):
        self.hdf_filename = hdf_filename
        self._file = self.open_file(self.hdf_filename, mode=mode)
        self.parent_name = parent_name
        self.mode = mode
        # Use the latest_data_name if one isn't provided
        if data_name is None:
            self.data_name = self.latest_data_name
        else:
            self.data_name = data_name
    
    def __str__(self):
        return self.parent_name + '-' + self.data_name

    def __repr__(self):
        fmt = '<TXMStore: {}/{}/{}>'
        fmt = fmt.format(self.hdf_filename, self.parent_name, self.data_name)
        return fmt
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        self.close()
    
    def frameset_names(self):
        """Returns a list of all the valid frameset representations."""
        names = []
        for key in self.data_group().keys():
            # Check if it's a "frameset" or not
            grp = self.data_group()
            context = grp[key].attrs.get('context', '')
            if context == 'frameset':
                names.append(key)
        return names        
    
    def map_names(self):
        """Returns a list of all the valid map representations."""
        names = []
        for key in self.data_group().keys():
            # Check if it's a "map" or not
            if self.data_group()[key].attrs.get('context','') == 'map':
                names.append(key)
        return names
    
    def open_file(self, filename, mode):
        return h5py.File(filename, mode=mode)
    
    def close(self):
        self._file.close()
    
    def data_tree(self):
        """Create a tree of the possible groups this store could access. The
        first level is samples, then data_groups (ie. same sample but
        different analysis status), then representations.
        """
        # Define a recursive function to walk the groups in the file
        def walk_groups(parent, level):
            # Recurse into children
            keys = getattr(parent, 'keys', lambda: [])
            datas = []
            for key in keys():
                # Check for whether this object is a frameset or a map
                node = parent[key]
                grpdata = {
                    'name': key,
                    'path': node.name,
                    'context': node.attrs.get('context', None),
                    'ndim': getattr(node, 'ndim', 0),
                    'level': level,
                    # Resurse
                    'children': walk_groups(parent=node, level=level+1),
                }
                datas.append(grpdata)
            return datas
        # Start the recursion at the top
        tree = walk_groups(self._file, level=0)
        return tree
    
    def fork_data_group(self, dest, src=None):
        """Turn on different active data group for this store. This method
        deletes the existing group and copies symlinks from the
        current one.
        
        """
        # Switch to the group given by `src`
        if src is not None:
            self.data_name = src
        # Check that the current and target groups are not the same
        if dest == self.data_name:
            log.critical('Refusing to fork group "%s" to itself', dest)
            msg = 'Refusing to fork myself to myself ({})'.format(dest)
            raise CreateGroupError(msg)
        log.info('Forking data group "%s" to "%s"', src, dest)
        # Delete the old group and overwrite it
        parent = self.parent_group()
        if dest in parent.keys():
            del parent[dest]
        # Copy the old one with symlinks
        new_group = parent.copy(self.data_group(), dest, shallow=True)
        self.latest_data_name = dest
        self.data_name = dest
        return new_group
    
    @property
    def latest_data_name(self):
        name = self.parent_group().attrs['latest_data_name']
        return name
    
    @latest_data_name.setter
    def latest_data_name(self, val):
        self.parent_group().attrs['latest_data_name'] = val
    
    def validate_parent_group(self, name):
        """Retrieve the real parent group name for a possible parent_group.
        
        If ``name`` is None and only one group exists in the file,
        then that group name will be returned. If ``name`` is in the
        file, then ``name`` will be returned. If ``name`` is not in
        the file, a GroupKeyError will be raised.
        """
        if name is None and len(self._file.keys()) == 1:
            new_name = list(self._file.keys())[0]
        elif name not in self._file.keys():
            raise GroupKeyError("Cannot load parent group '{group}'. "
                                "Valid choices are {choices}."
                                "".format(group=name, choices=list(self._file.keys())))
        else:
            new_name = name
        return new_name
    
    def parent_group(self):
        """Retrieve the top-level HDF5 group object for this file and
        groupname."""
        try:
            parent_group = self.validate_parent_group(self.parent_name)
            group = self._file[parent_group]
        except (TypeError, KeyError):
            # Invalid group name, throw an exception
            msg = 'Cannot load parent group "{group}". Valid choices are {choices}'
            try:
                choices = list(self._file.keys())
            except:
                choices = 'unavailable'
            msg = msg.format(group=self.parent_name, choices=choices)
            raise GroupKeyError(msg) from None
        return group
    
    def data_group(self):
        """Retrieve the currently active second-level HDF5 group object for
        this file and groupname. Ex. "imported" or "aligned_frames".
        """
        if self.data_name not in self.parent_group().keys():
            msg = "Group {} does not exists. Run TXMStore.fork_data_group('{}') first"
            raise CreateGroupError(msg.format(self.data_name, self.data_name))
        return self.parent_group()[self.data_name]
   
    def replace_dataset(self, name, data, context=None, attrs={},
                        compression=None, *args, **kwargs):
        """Wrapper for h5py.create_dataset that removes the existing dataset
        if it exists.
        
        Parameters
        ----------
        name : str
          HDF5 groupname name to give this dataset.
        data : np.ndarray
          Numpy array of data to be saved.
        context : str, optional
          Specifies what kind of data is stored. Eg. "frameset",
          "metadata", "map".
        attrs : dict, optional
          Dictionary containing HDF5 metadata attributes to be set on
          the resulting dataset.
        compression : str, optional
          What type of compression to use. See HDF5 documentation for
          options.
        *args
          Arguments to pass to h5py's ``create_dataset`` method.
        **kwargs
          Keyword arguments to pass to h5py's ``create_dataset`` method.
        
        """
        # Remove the existing dataset if possible
        try:
            attrs = self.data_group()[name].attrs
            del self.data_group()[name]
        except KeyError as e:
            pass
        # Perform the actual group creation
        ds = self.data_group().create_dataset(name=name, data=data,
                                              compression=compression,
                                              *args, **kwargs)
        # Set metadata attributes
        if context is not None:
            ds.attrs['context'] = context
        for key, val in attrs.items():
            ds.attrs[key] = val
        return ds
    
    def get_dataset(self, name):
        """Attempt to open the requested dataset.
        
        Parameters
        ==========
        name : str
          The name of the dataset to open in the data group.
        
        Returns
        -------
        data : hyp5.Dataset
          An open HDF5 dataset
        
        Raises
        ------
        exceptions.GroupKeyError
          If the dataset does not exist in the file.
        
        """
        # Check for some bad dataset names
        if name is None:
            msg = "dataset `None` not found in file '{}'"
            msg = msg.format(self.hdf_filename)
            raise GroupKeyError(msg)
        elif name not in self.data_group().keys():
            msg = "dataset '{}' not found in group '{}' file '{}'"
            msg = msg.format(name, self.data_group().name, self.hdf_filename)
            raise GroupKeyError(msg)
        else:
            data = self.data_group()[name]
        return data
    
    def has_dataset(self, name):
        """Return a boolean indicated whether this dataset exists in the HDF
        file.
        
        """
        try:
            result = name in self.data_group().keys()
        except TypeError:
            result = False
        return result
    
    def frame_source(self, name):
        """Get the name of the frames that went into creating a map."""
        attrs = getattr(self.get_dataset(name), 'attrs', {})
        source = attrs.get('frame_source', name)
        return source
    
    def get_frames(self, name):
        """Return the source frame data for the given data name.
        
        This is similar to ``get_dataset`` except that if the
        data are a map, then get the frames that went into making it.
        
        Parameters
        ----------
        name : str
          The dataset name for which to retrieve frames.
        
        Returns
        -------
        dataset : h5py.Dataset
          The requested frameset. If the dataset is actually a map, as
          determined by the "context" attribute, then the related
          frame source attribute will be retrieved.
        
        """
        dataset = self.get_dataset(name)
        # If it's a map, then return the source frames instead
        if dataset.attrs.get('context', None) == 'map':
            try:
                dataset = self.get_dataset(dataset.attrs['frame_source'])
            except KeyError:
                source_desc = dataset.attrs.get('frame_source', 'None')
                # raise FrameSourceError(
                log.warning(
                    "Invalid frame source {} specified for group {}"
                    "".format(source_desc, self.data_name))
                dataset = self.get_dataset('optical_depths')
        return dataset
    
    @property
    def relative_positions(self):
    
        """(x, y, z) position values for each frame."""
        return self.data_group()['relative_positions']
    
    @relative_positions.setter
    def relative_positions(self, val):
        self.replace_dataset('relative_positions', val, context='metadata')
        self.data_group()['relative_positions'].attrs['order'] = "(x, y, z)"
   
    @property
    def pixel_unit(self):
        return self.data_group()['pixel_sizes'].attrs['unit']
    
    @pixel_unit.setter
    def pixel_unit(self, val):
        self.data_group()['pixel_sizes'].attrs['unit'] = val
    
    @property
    def signal_method(self):
        """String describing how the previously extracted signals were
        calculated.
        """
        return self.data_group()['extracted_signals'].attrs['method']
    
    @signal_method.setter
    def signal_method(self, val):
        self.data_group()['extracted_signals'].attrs['method'] = val
    
    @property
    def timestamps(self):
        return self.data_group()['timestamps']
    
    @timestamps.setter
    def timestamps(self, val):
        # S32 is the 32-character ACSII string type for numpy
        val = np.array(val, dtype="S32")
        self.replace_dataset('timestamps', val, dtype="S32", context='metadata',
                             attrs={'timezone', "UTC"})
    
    @property
    def filenames(self):
        return self.data_group()['filenames']
    
    @filenames.setter
    def filenames(self, val):
        # S100 is the 100-character ACSII string type for numpy
        val = np.array(val, dtype="S100")
        self.replace_dataset('filenames', val, dtype="S100", context='metadata')
    
    @property
    def fit_parameters(self):
        return self.get_dataset('fit_parameters')
    
    @fit_parameters.setter
    def fit_parameters(self, val):
        attrs = {
            'parameters': str(xm.kedge_params),
        }
        return self.replace_dataset('fit_parameters', val,
                                    attrs=attrs, context="metadata",
                                    dtype=np.float64)
